/*--------------------------------------------------
   TGB Dual - Gameboy Emulator -
   Copyright (C) 2001  Hii

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//------------------------------------------------
// CPU j[jbNȊO (I/OIRQ )

#include "gb.h"
#include "../string.h"
#include "_bit_table.h"

#define Z_FLAG 0x40
#define H_FLAG 0x10
#define N_FLAG 0x02
#define C_FLAG 0x01

/////////////////////////////////////////
//
// oϐ
//
/////////////////////////////////////////

//structgp΂gȂ݂Ȃ̂ŁAʂ̕ϐɕύXB֘ȀC߂Ȃ̂ŒӁB - LCK
struct cpu_regs _c_regs;
word c_regs_AF;
word c_regs_BC;
word c_regs_DE;
word c_regs_HL;
word c_regs_SP;
word c_regs_PC;
byte c_regs_I;

byte ram[0x2000*4];
byte vram[0x2000*2];
byte stack[0x80];
byte oam[0xA0];
byte spare_oam[0x18];
byte ext_mem[16];

byte *vram_bank;
byte *ram_bank;

byte z802gb[256],gb2z80[256];
dword rp_que[256];
int que_cur;
//	word org_pal[16][4];
int total_clock,rest_clock,sys_clock,seri_occer,div_clock;
bool halt,speed,speed_change,dma_executing;
bool b_trace;
int dma_src;
int dma_dest;
int dma_rest;
int gdma_rest;
bool b_dma_first;

int last_int;
volatile bool int_disable_next;
volatile bool int_invoke_next;

byte *dma_src_bank;
byte *dma_dest_bank;

byte _ff6c,_ff72,_ff73,_ff74,_ff75;

int rp_fd = -1;
unsigned long rp_time;
byte rp_data;

/////////////////////////////////////////
//
// o֐
//
/////////////////////////////////////////
void cpu_init(void)
{
	int i;

	b_trace=false;

	for (i=0;i<256;i++){
		z802gb[i]=((i&0x40)?0x80:0)|((i&0x10)?0x20:0)|((i&0x02)?0x40:0)|((i&0x01)?0x10:0);
		gb2z80[i]=((i&0x80)?0x40:0)|((i&0x40)?0x02:0)|((i&0x20)?0x10:0)|((i&0x10)?0x01:0);
	}
	
	cpu_reset();
}

void cpu_reset(void)
{
	c_regs_AF=(rom_get_info()->gb_type>=3)?0x11b0:0x01b0;
	c_regs_BC=(rom_get_info()->gb_type>=4)?0x0113:0x0013;
	c_regs_DE=0x00D8;
	c_regs_HL=0x014D;
	c_regs_I=0;
	c_regs_SP=0xFFFE;
	c_regs_PC=0x100;

	vram_bank=vram;
	ram_bank=ram+0x1000;

	rest_clock=0;
	total_clock=sys_clock=div_clock=0;
	seri_occer=0x7fffffff;
	halt=false;
	speed=false;
	speed_change=false;
	dma_executing=false;
	b_dma_first=false;
	gdma_rest=0;

	last_int=0;
	int_disable_next=false;
	int_invoke_next=false;

	_memset(ram,0,sizeof(ram));
	_memset(vram,0,sizeof(vram));
	_memset(stack,0,sizeof(stack));
	_memset(oam,0,sizeof(oam));
	_memset(spare_oam,0,sizeof(spare_oam));

	rp_que[0]=0x000001cc;
	rp_que[1]=0x00000000;
	que_cur=1;
	
	if(rp_fd >= 0){
		sceIoClose(rp_fd);
		rp_fd = -1;
	}
	rp_time = 0;
}

void cpu_save_state(int *dat)
{
	dat[0]=(ram_bank-ram)/0x1000;
	dat[1]=(vram_bank-vram)/0x2000;

	dat[2]=(speed?1:0);
	dat[3]=(dma_executing?1:0);
	dat[4]=dma_src;
	dat[5]=dma_dest;
	dat[6]=dma_rest;
	dat[7]=(speed_change?1:0);
}

void cpu_save_state_ex(int *dat)
{
	dat[0]=div_clock;
	dat[1]=rest_clock;
	dat[2]=sys_clock;
	dat[3]=total_clock;
}

#define _USE_CPU_IO_JMPTBL
#ifdef  _USE_CPU_IO_JMPTBL
#include "cpu_io.h"
#else  // _USE_CPU_IO_JMPTBL

inline byte cpu_io_read(word adr)
{
	int tmp;
	byte ret;

	switch(adr){
	case 0xFF00://P1(pbh)
		if(now_gb_mode==2){
			sgb_readingcontrol |= 4;
			sgb_reset_state();
			/*
			if(sgb_multiplayer){
				switch(sgb_nextcontrol)
				{
				case 0x0F:
					//Check_KBInput(0);
				break;
				case 0x0E:
					//Check_KBInput(1);
				break;
				case 0x0D:
					//Check_KBInput(2);
				break;
				case 0x0C:
					//Check_KBInput(3);
				break;
				}
			}
			*/
		}
		
		tmp=pad_state;
		if (g_regs.P1==0x03)
			return 0xff;
		switch((g_regs.P1>>4)&0x3){
		case 0:
			return 0xC0|((tmp&0x81?0:1)|(tmp&0x42?0:2)|(tmp&0x24?0:4)|(tmp&0x18?0:8));
		case 1:
			return 0xD0|((tmp&0x01?0:1)|(tmp&0x02?0:2)|(tmp&0x04?0:4)|(tmp&0x08?0:8));
		case 2:
			return 0xE0|((tmp&0x80?0:1)|(tmp&0x40?0:2)|(tmp&0x20?0:4)|(tmp&0x10?0:8));
		case 3:
			if(now_gb_mode==2 && sgb_multiplayer) 
				return 0xF0 | sgb_nextcontrol;
			else
				return 0xFF;
		}
		return 0x00;
	case 0xFF01://SB(VAʐMM)
//		fprintf(file,"Read SB %02X\n",g_regs.SB);
		return g_regs.SB;
	case 0xFF02://SC(VARg[)
//		fprintf(file,"Read SC %02X\n",g_regs.SC);
		return (g_regs.SC&0x83)|0x7C;
	case 0xFF04://DIV(fBoC_[?)
		return g_regs.DIV;
	case 0xFF05://TIMA(^C}JE^)
		return g_regs.TIMA;
	case 0xFF06://TMA(^C})
		return g_regs.TMA;
	case 0xFF07://TAC(^C}Rg[)
		return g_regs.TAC;
	case 0xFF0F://IF(肱݃tO)
		return g_regs.IF;
	case 0xFF40://LCDC(LCDRg[)
		return g_regs.LCDC;
	case 0xFF41://STAT(LCDXe[^X)
		return g_regs.STAT|0x80;
	case 0xFF42://SCY(XN[Y)
		return g_regs.SCY;
	case 0xFF43://SCX(XN[X)
		return g_regs.SCX;
	case 0xFF44://LY(LCDC YW)
		return g_regs.LY;
	case 0xFF45://LYC(LYr)
		return g_regs.LYC;
	case 0xFF46://DMA(DMA])
		return 0;
	case 0xFF47://BGP(wipbg)
		return g_regs.BGP;
	case 0xFF48://OBP1(IuWFNgpbg1)
		return g_regs.OBP1;
	case 0xFF49://OBP2(IuWFNgpbg2)
		return g_regs.OBP2;
	case 0xFF4A://WY(EChEYW)
		return g_regs.WY;
	case 0xFF4B://WX(EChEXW)
		return g_regs.WX;

		//ȉJ[ł̒ǉ
	case 0xFF4D://KEY1VXeNbNύX
		return (speed?0x80:(cg_regs.KEY1&1)?1:0x7E);
	case 0xFF4F://VBK(VRAMoN؂ւ)
		return cg_regs.VBK;
	case 0xFF51://HDMA1(])
		return dma_src>>8;
	case 0xFF52://HDMA2(])
		return dma_src&0xff;
	case 0xFF53://HDMA3(])
		return dma_dest>>8;
	case 0xFF54://HDMA4(]扺)
		return dma_dest&0xff;
	case 0xFF55://HDMA5(]s)
		return (dma_executing?((dma_rest-1)&0x7f):0xFF);
	case 0xFF56://RP(ԊO)
		{
			int bit, size;
			unsigned long cur_time;
			if ((cg_regs.RP&0xC0)==0xC0){
				cur_time = sceKernelLibcClock();
				while(cur_time-rp_time >= 25000){
					size = sceIoRead(rp_fd, &rp_data, 1);
					if(size==0){
						rp_time = cur_time-25000;
						break;
					}else{
						rp_time += 25000;
					}
				}
				bit = rp_data >> ( (cur_time-rp_time)/3125 ) & 1;
			}else{
				bit = 0;
			}
			return cg_regs.RP & 0xFD | ((bit^1)<<1);
		}
	case 0xFF68://BCPS(BGpbgݎw)
		return cg_regs.BCPS;
	case 0xFF69://BCPD(BGpbg݃f[^)
		if (cg_regs.BCPS&1)
			ret=lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3]>>8;
		else
			ret=lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3]&0xff;
/*		if (cg_regs.BCPS&1)
			ret=renderer_unmap_color(lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3])>>8;
		else
			ret=renderer_unmap_color(lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3])&0xff;
*/		//|C^̓CNgȂ()
		return ret;
	case 0xFF6A://OCPS(OBJpbgݎw)
		return cg_regs.OCPS;
	case 0xFF6B://OCPD(OBJpbg݃f[^)
		if (cg_regs.OCPS&1)
			ret=lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3]>>8;
		else
			ret=lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3]&0xff;
/*		if (cg_regs.OCPS&1)
			ret=renderer_unmap_color(lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3])>>8;
		else
			ret=renderer_unmap_color(lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3])&0xff;
*/		return ret;
	case 0xFF70://SVBK(RAMoN؂ւ)
		return cg_regs.SVBK;

	case 0xFFFF://IE(肱݃}XN)
		return g_regs.IE;

	// undocumented register
	case 0xFF6C:
		return _ff6c&1;
	case 0xFF72:
		return _ff72;
	case 0xFF73:
		return _ff73;
	case 0xFF74:
		return _ff74;
	case 0xFF75:
		return _ff75&0x70;
	case 0xFF76:
		return 0;
	case 0xFF77:
		return 0;
	default:
		if (adr>0xFF0F&&adr<0xFF40){
			return apu_read(adr);
		}else if ((adr>0xff70)&&(adr<0xff80))
			return ext_mem[adr-0xff71];
		else
			return 0;
	}
}

inline void cpu_io_write(word adr,byte dat)
{
	word tmp_adr;

	switch(adr){
	case 0xFF00://P1(pbh)
		if(now_gb_mode==2) sgb_transfer(dat);
//		g_regs.P1=dat;
		g_regs.P1 = ((g_regs.P1 & 0xCF)|(dat&0x30));
		return;
	case 0xFF01://SB(VAVAʐMM)
		g_regs.SB=dat;
		return;
	case 0xFF02://SC(Rg[)
		if (rom_get_info()->gb_type<=2){
			g_regs.SC=dat&0x81;
			if ((dat&0x80)&&(dat&1)) // MJn
				seri_occer=total_clock+512;
		}
		else{ // GBCł̊g
			g_regs.SC=dat&0x83;
			if ((dat&0x80)&&(dat&1)) // MJn
				if (dat&2)
					seri_occer=total_clock+512*8/32; // ]xʏ32{
				else
					seri_occer=total_clock+512*8;
		}
		return;
	case 0xFF04://DIV(fBoC_[)
		g_regs.DIV=0;
		return;
	case 0xFF05://TIMA(^C}JE^)
		g_regs.TIMA=dat;
//			sys_clock=0;
		return;
	case 0xFF06://TMA(^C})
		g_regs.TMA=dat;
//			sys_clock=0;
		return;
	case 0xFF07://TAC(^C}Rg[)
		if ((dat&0x04)&&!(g_regs.TAC&0x04))
			sys_clock=0;
		g_regs.TAC=dat;
		return;
	case 0xFF0F://IF(肱݃tO)
		g_regs.IF=dat;
		cpu_irq_check();
		return;
	case 0xFF40://LCDC(LCDRg[)
		if ((dat&0x80)&&(!(g_regs.LCDC&0x80))){
			g_regs.LY=0;
			lcd_clear_win_count();
		}
		g_regs.LCDC=dat;
//			fprintf(file,"LCDC=%02X at line %d\n",dat,g_regs.LY);
		return;
	case 0xFF41://STAT(LCDXe[^X)
//		if (rom_get_info()->gb_type==1) // IWiGBɂĂ̂悤ȌۂN炵
		if (rom_get_info()->gb_type<=2)
			if (!(g_regs.STAT&0x02))
				g_regs.IF|=INT_LCDC;

		g_regs.STAT=(g_regs.STAT&0x7)|(dat&0x78);
		cpu_irq_check();
		return;
	case 0xFF42://SCY(XN[Y)
		g_regs.SCY=dat;
		return;
	case 0xFF43://SCX(XN[X)
		g_regs.SCX=dat;
		return;
	case 0xFF44://LY(LCDC YW)
//			g_regs.LY=0;
		lcd_clear_win_count();
		return;
	case 0xFF45://LYC(LYr)
		g_regs.LYC=dat;
		return;
	case 0xFF46://DMA(DMA])
		switch(dat>>5){
		case 0:
		case 1:
			_memcpy(oam,get_rom()+dat*256,0xA0);
			break;
		case 2:
		case 3:
			_memcpy(oam,mbc_get_rom()+dat*256,0xA0);
			break;
		case 4:
			_memcpy(oam,vram_bank+(dat&0x1F)*256,0xA0);
			break;
		case 5:
			_memcpy(oam,mbc_get_sram()+(dat&0x1F)*256,0xA0);
			break;
		case 6:
			if (dat&0x10)
				_memcpy(oam,ram_bank+(dat&0x0F)*256,0xA0);
			else
				_memcpy(oam,ram+(dat&0x0F)*256,0xA0);
			break;
		case 7:
			if (dat<0xF2){
				if (dat&0x10)
					_memcpy(oam,ram_bank+(dat&0x0F)*256,0xA0);
				else
					_memcpy(oam,ram+(dat&0x0F)*256,0xA0);
			}
			break;
		}
		return;
	case 0xFF47://BGP(wipbg)
		g_regs.BGP=dat;
		return;
	case 0xFF48://OBP1(IuWFNgpbg1)
		g_regs.OBP1=dat;
		return;
	case 0xFF49://OBP2(IuWFNgpbg2)
		g_regs.OBP2=dat;
		return;
	case 0xFF4A://WY(EChEYW)
		g_regs.WY=dat;
		return;
	case 0xFF4B://WX(EChEXW)
		g_regs.WX=dat;
		return;

		//ȉJ[ł̒ǉ
	case 0xFF4D://KEY1VXeNbNύX
//			speed=dat&1;
		cg_regs.KEY1=dat&1;
		speed_change=dat&1;
		return;
	case 0xFF4F://VBK(VRAMoN؂ւ)
		if (dma_executing)
			return;
		vram_bank=vram+0x2000*(dat&0x01);
		cg_regs.VBK=dat;//&0x01;
		return;
	case 0xFF51://HDMA1(])
		dma_src&=0x00F0;
		dma_src|=(dat<<8);
		return;
	case 0xFF52://HDMA2(])
		dma_src&=0xFF00;
		dma_src|=(dat&0xF0);
		return;
	case 0xFF53://HDMA3(])
		dma_dest&=0x00F0;
		dma_dest|=((dat&0xFF)<<8);
		return;
	case 0xFF54://HDMA4(]扺)
		dma_dest&=0xFF00;
		dma_dest|=(dat&0xF0);
		return;
	case 0xFF55://HDMA5(]s)
		tmp_adr=0x8000+(dma_dest&0x1ff0);
//			fprintf(file,"%03d : %04X -> %04X  %d byte %s\n",g_regs.LY,dma_src,dma_dest,((dat&0x7f)+1)*16,(dat&0x80)?"delay":"immidiately");
		if ((dma_src>=0x8000&&dma_src<0xA000)||(dma_src>=0xE000)||(!(tmp_adr>=0x8000&&tmp_adr<0xA000))){
			cg_regs.HDMA5=0;
			return;
		}
		if (dat&0x80){ //HBlank
			if (dma_executing){
				dma_executing=false;
				dma_rest=0;
				cg_regs.HDMA5=0xFF;
				return;
			}
			dma_executing=true;
			b_dma_first=true;
			dma_rest=(dat&0x7F)+1;
			cg_regs.HDMA5=0;
/*				dma_dest_bank=vram_bank;
			if (dma_src<0x4000)
				dma_src_bank=get_rom();
			else if (dma_src<0x8000)
				dma_src_bank=mbc_get_rom()-0x4000;
			else if (dma_src>=0xA000&&dma_src<0xC000)
				dma_src_bank=mbc_get_sram()-0xA000;
			else if (dma_src>=0xC000&&dma_src<0xD000)
				dma_src_bank=ram-0xC000;
			else if (dma_src>=0xD000&&dma_src<0xE000)
				dma_src_bank=ram_bank-0xD000;
			else dma_src_bank=NULL;
*/			}
		else{ //ʏDMA
			if (dma_executing){
				dma_executing=false;
				dma_rest=0;
				cg_regs.HDMA5=0xFF;
//					fprintf(file,"dma stopped\n");
				return;
			}
			// ǂHBlankȊOȂ炢łOK݂
//				if (!(((g_regs.STAT&3)==1)||(!(g_regs.LCDC&0x80)))){
//					cg_regs.HDMA5=0;
//					return;
//				}

			dma_executing=false;
			dma_rest=0;
			cg_regs.HDMA5=0xFF;

			switch(dma_src>>13){
			case 0:
			case 1:
				_memcpy(vram_bank+(dma_dest&0x1ff0),get_rom()+(dma_src),16*(dat&0x7F)+16);
				break;
			case 2:
			case 3:
				_memcpy(vram_bank+(dma_dest&0x1ff0),mbc_get_rom()+(dma_src),16*(dat&0x7F)+16);
				break;
			case 4:
				break;
			case 5:
				_memcpy(vram_bank+(dma_dest&0x1ff0),mbc_get_sram()+(dma_src&0x1FFF),16*(dat&0x7F)+16);
				break;
			case 6:
				if (dma_src&0x1000)
					_memcpy(vram_bank+(dma_dest&0x1ff0),ram_bank+(dma_src&0x0FFF),16*(dat&0x7F)+16);
				else
					_memcpy(vram_bank+(dma_dest&0x1ff0),ram+(dma_src&0x0FFF),16*(dat&0x7F)+16);
				break;
			case 7:
				break;
			}
			dma_src+=((dat&0x7F)+1)*16;
			dma_dest+=((dat&0x7F)+1)*16;

			gdma_rest=456*2+((dat&0x7f)+1)*32*(speed?2:1); // CPU p[
		}
		return;
	case 0xFF56://RP(ԊO)
		if ((dat&0xC0)==0xC0){
			if (rp_fd<0){
				rp_fd = sceIoOpen("irda0:",O_RDWR, 777);
				rp_data = sceIoRead(rp_fd, &rp_data, 1);
				rp_time = sceKernelLibcClock();
			}
		}else{
			if (rp_fd>=0){
				sceIoClose(rp_fd);
				rp_fd = -1;
			}
			rp_time = 0;
		}
		cg_regs.RP=dat;
		
		return;
	case 0xFF68://BCPS(BGpbgݎw)
		cg_regs.BCPS=dat;
		return;
	case 0xFF69://BCPD(BGpbg݃f[^ xBBBBBGG GGGRRRRR)
		if (cg_regs.BCPS&1){
			lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3]=
			(lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3]&0xff)|(dat<<8);
		}
		else{
			lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3]=
			(lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3]&0xff00)|dat;
		}
/*			if (cg_regs.BCPS&1){
			lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3]=
				renderer_map_color(((renderer_unmap_color(lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3])&0xff)|(dat<<8)));
		}
		else{
			lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3]=
				renderer_map_color(((renderer_unmap_color(lcd_get_pal((cg_regs.BCPS>>3)&7)[(cg_regs.BCPS>>1)&3])&0xff00)|(dat)));
		}*/
		cg_regs.BCPD=dat;
		if (cg_regs.BCPS&0x80)
			cg_regs.BCPS=0x80|((cg_regs.BCPS+1)&0x3f);
//			fprintf(file,"%d :BCPS = %02X\n",g_regs.LY,dat);
		return;
	case 0xFF6A://OCPS(OBJpbgݎw)
		cg_regs.OCPS=dat;
		return;
	case 0xFF6B://OCPD(OBJpbg݃f[^)
		if (cg_regs.OCPS&1){
			lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3]=
			(lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3]&0xff)|(dat<<8);
		}
		else{
			lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3]=
			(lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3]&0xff00)|dat;
		}
/*			if (cg_regs.OCPS&1){
			lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3]=
				renderer_map_color(((renderer_unmap_color(lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3])&0xff)|(dat<<8)));
		}
		else{
			lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3]=
				renderer_map_color(((renderer_unmap_color(lcd_get_pal(((cg_regs.OCPS>>3)&7)+8)[(cg_regs.OCPS>>1)&3])&0xff00)|(dat)));
		}*/
		cg_regs.OCPD=dat;
		if (cg_regs.OCPS&0x80)
			cg_regs.OCPS=0x80|((cg_regs.OCPS+1)&0x3f);
		return;
	case 0xFF70://SVBK(RAMoN؂ւ)
//			if (dma_executing)
//				return;

		dat=(!(dat&7))?1:(dat&7);
		cg_regs.SVBK=dat;
		ram_bank=ram+0x1000*dat;
		return;

	case 0xFFFF://IE(肱݃}XN)
		g_regs.IE=dat;
//			g_regs.IF=0;
//			fprintf(file,"IE = %02X\n",dat);
		cpu_irq_check();
		return;

	// undocumented register
	case 0xFF6C:
		_ff6c=dat&1;
		return;
	case 0xFF72:
		_ff72=dat;
		return;
	case 0xFF73:
		_ff73=dat;
		return;
	case 0xFF74:
		_ff74=dat;
		return;
	case 0xff75:
		_ff75=dat&0x70;
		return;

	default:
		if (adr>0xFF0F&&adr<0xFF40){
			apu_write(adr,dat,total_clock);
			return;
		}else if ((adr>0xff70)&&(adr<0xff80))
			ext_mem[adr-0xff71]=dat;
	}
}

#endif //_USE_CPU_IO_JMPTBL


void cpu_restore_state(int *dat)
{
	ram_bank=ram+dat[0]*0x1000;
	vram_bank=vram+dat[1]*0x2000;

	speed=(dat[2]?true:false);
	dma_executing=(dat[3]?true:false);
	dma_src=dat[4];
	dma_dest=dat[5];
	dma_rest=dat[6];
	speed_change=(dat[7]?true:false);
}

void cpu_restore_state_ex(int *dat)
{
	div_clock=dat[0];
	rest_clock=dat[1];
	sys_clock=dat[2];
	total_clock=dat[3];
}

//inline byte cpu_read_direct(word adr)
byte cpu_read_direct_ord(word adr)
{
	switch(adr>>13){
	case 0:
	case 1:
		return get_rom()[adr];//ROM̈
	case 2:
	case 3:
		//return get_rom()[adr];//oN\ROM
		return mbc_get_rom()[adr];//oN\ROM
	case 4:
		return vram_bank[adr&0x1FFF];//8KBVRAM
	case 5:
		if (mbc_is_ext_ram())
			return mbc_get_sram()[adr&0x1FFF];//J[gbWRAM
		else
			return mbc_ext_read(adr);
	case 6:
		if (adr&0x1000)
			return ram_bank[adr&0x0fff];
		else
			return ram[adr&0x0fff];
	case 7:
		if (adr<0xFE00){
			if (adr&0x1000)
				return ram_bank[adr&0x0fff];
			else
				return ram[adr&0x0fff];
		}
		else if (adr<0xFEA0)
			return oam[adr-0xFE00];//object attribute memory
		else if (adr<0xFF00)
			return spare_oam[(((adr-0xFFA0)>>5)<<3)|(adr&7)];

#ifdef _USE_CPU_IO_JMPTBL
		else
			return cpu_io_read_JmpTbl[ adr & 0x00FF ]( adr ) ;
#else  //_USE_CPU_IO_JMPTBL
		else if (adr<0xFF80)
			return cpu_io_read(adr);//I/O
		else if (adr<0xFFFF)
			return stack[adr-0xFF80];//stack
		else
			return cpu_io_read(adr);//I/O
#endif //_USE_CPU_IO_JMPTBL

	}
	return 0;
}

//ZăCCɂԂ - LCK
inline byte cpu_read_direct(word adr)
{
	if ((adr&0x8000)==0) {
		return ((adr&0x4000)==0)?get_rom()[adr]:mbc_get_rom()[adr];
	} else if ((adr&0xe000)==0xc000) {
		return ((adr&0x1000)==0)?ram[adr&0xfff]:ram_bank[adr&0xfff];
//	} else if ((adr&0xe000)==0xa000) {
//		return (mbc_is_ext_ram())?mbc_get_sram()[adr&0x1FFF]:mbc_ext_read(adr);
	}
	return cpu_read_direct_ord(adr);
}

inline byte read(word adr)
{
	return cpu_read_direct(adr);
}

inline word readw(word adr)
{
	return read(adr)|(read(adr+1)<<8);
}

inline void writew(word adr,word dat)
{
	write(adr,(byte)dat);
	write(adr+1,dat>>8);
}

inline byte op_read()
{
	return read(c_regs_PC++);
}
/*
inline word op_readw()
{
	c_regs.PC+=2;
	return readw(c_regs.PC-2);
}
*/
/*
//̂قƎv - LCK
inline byte op_read()
{
	word pc=c_regs_PC;
	byte r=read(pc);
	c_regs_PC=pc+1;
	return r;
}*/

//̂قƎv - LCK
inline word op_readw()
{
	word r=readw(c_regs_PC);
	c_regs_PC+=2;
	return r;
}

//inline void write(word adr,byte dat)
void write_ord(word adr,byte dat)
{
	switch(adr>>13){
	case 0:
	case 1:
	case 2:
	case 3:
		mbc_write(adr,dat);
		break;
	case 4:
		vram_bank[adr&0x1FFF]=dat;
		break;
	case 5:
		if (mbc_is_ext_ram())
			mbc_get_sram()[adr&0x1FFF]=dat;//J[gbWRAM
		else
			mbc_ext_write(adr,dat);
		break;
	case 6:
		if (adr&0x1000)
			ram_bank[adr&0x0fff]=dat;
		else
			ram[adr&0x0fff]=dat;
		break;
	case 7:
		if (adr<0xFE00){
			if (adr&0x1000)
				ram_bank[adr&0x0fff]=dat;
			else
				ram[adr&0x0fff]=dat;
		}
		else if (adr<0xFEA0)
			oam[adr-0xFE00]=dat;
		else if (adr<0xFF00)
			spare_oam[(((adr-0xFFA0)>>5)<<3)|(adr&7)]=dat;

#ifdef _USE_CPU_IO_JMPTBL
		else
			cpu_io_write_JmpTbl[ adr & 0x00FF ]( adr, dat ) ;
#else  //_USE_CPU_IO_JMPTBL
		else if (adr<0xFF80)
			cpu_io_write(adr,dat);//I/O
		else if (adr<0xFFFF)
			stack[adr-0xFF80]=dat;//stack
		else
			cpu_io_write(adr,dat);//I/O
#endif //_USE_CPU_IO_JMPTBL
		break;
	}
}

//ZăCCɂԂ - LCK
inline void write(word adr,byte dat)
{
	if ((adr&0xe000)==0xc000) {
		if (adr&0x1000)
			ram_bank[adr&0x0fff]=dat;
		else
			ram[adr&0x0fff]=dat;
	} else if ((adr&0xe000)==0xa000) {
		if (mbc_is_ext_ram())
			mbc_get_sram()[adr&0x1FFF]=dat;//J[gbWRAM
		else
			mbc_ext_write(adr,dat);
	} else {
		write_ord(adr,dat);
	}
}

/*
static int cycles[256] =
{
	4,12,8,8,4,4,8,4,4,12,8,8,4,4,8,4,
	8,12,8,8,4,4,8,4,8,12,8,8,4,4,8,4,
	8,12,8,8,4,4,8,4,8,12,8,8,4,4,8,4,
	8,12,8,8,12,12,12,4,8,12,8,8,4,4,8,4,
	4,4,4,4,4,4,8,4,4,4,4,4,4,4,8,4,
	4,4,4,4,4,4,8,4,4,4,4,4,4,4,8,4,
	4,4,4,4,4,4,8,4,4,4,4,4,4,4,8,4,
	8,8,8,8,8,8,4,8,4,4,4,4,4,4,8,4,
	4,4,4,4,4,4,8,4,4,4,4,4,4,4,8,4,
	4,4,4,4,4,4,8,4,4,4,4,4,4,4,8,4,
	4,4,4,4,4,4,8,4,4,4,4,4,4,4,8,4,
	4,4,4,4,4,4,8,4,4,4,4,4,4,4,8,4,
	8,12,16,16,12,16,8,16,8,16,16,0,12,24,8,16,
	8,12,16,16,12,16,8,16,8,16,16,16,12,24,8,16,
	12,12,8,12,12,16,8,16,16,4,16,12,12,12,8,16,
	12,12,8,4,12,16,8,16,12,8,16,4,12,12,8,16
};
*/

static char cycles[256] =
{
//   0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
	 4,12, 8, 8, 4, 4, 8, 4,20, 8, 8, 8, 4, 4, 8, 4,//0
	 4,12, 8, 8, 4, 4, 8, 4,12, 8, 8, 8, 4, 4, 8, 4,//1
	 8,12, 8, 8, 4, 4, 8, 4, 8, 8, 8, 8, 4, 4, 8, 4,//2
	 8,12, 8, 8,12,12,12, 4, 8, 8, 8, 8, 4, 4, 8, 4,//3
	 4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,//4
	 4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,//5
	 4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,//6
	 8, 8, 8, 8, 8, 8, 4, 8, 4, 4, 4, 4, 4, 4, 8, 4,//7
	 4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,//8
	 4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,//9
	 4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,//A
	 4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,//B
	 8,12,12,16,12,16, 8,16, 8,16,12, 0,12,24, 8,16,//C
	 8,12,12, 0,12,16, 8,16, 8,16,12, 0,12, 0, 8,16,//D
	12,12, 8, 0, 0,16, 8,16,16, 4,16, 0, 0, 0, 8,16,//E
	12,12, 8, 4, 0,16, 8,16,12, 8,16, 4, 0, 0, 8,16 //F
};

/*
static int cycles[256] =
{
   4,12, 8, 8, 4, 4, 8, 4,20, 8, 8, 8, 4, 4, 8, 4,
   4,12, 8, 8, 4, 4, 8, 4, 8, 8, 8, 8, 4, 4, 8, 4,
   8,12, 8, 8, 4, 4, 8, 4, 8, 8, 8, 8, 4, 4, 8, 4,
   8,12, 8, 8,12,12,12, 4, 8, 8, 8, 8, 4, 4, 8, 4,
   4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
   4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
   4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
   8, 8, 8, 8, 8, 8, 4, 8, 4, 4, 4, 4, 4, 4, 8, 4,
   4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
   4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
   4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
   4, 4, 4, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4, 4, 8, 4,
   8,12,12,12,12,16, 8,32, 8, 8,12, 0,12,12, 8,32,
   8,12,12, 0,12,16, 8,32, 8, 8,12, 0,12, 0, 8,32,
  12,12, 8, 0, 0,16, 8,32,16, 4,16, 0, 0, 0, 8,32,
  12,12, 8, 4, 0,16, 8,32,12, 8,16, 4, 0, 0, 8,32
};
*/

static char cycles_cb[256] =
{
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
   8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
   8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
   8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8
};

/*static int cycles_cb[256] =
{
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8,
   8, 8, 8, 8, 8, 8,16, 8, 8, 8, 8, 8, 8, 8,16, 8
};
*/
/*
static const byte ZTable[256] =
{
  Z_FLAG,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
};*/


//#define Z_FLAG 0x40
//d==0̎0x40ԂBȊO͂OԂB
static inline byte GenZF(byte d)
{
	byte ret;
	asm (
		"		sltiu	%0,%1,1 "				"\n"
		"		sll		%0,%0,6 "				"\n"
			:	"=&r" (ret)		// %0
			:	"r" (d)			// %1
	);
	return ret;
}


byte cpu_seri_send(byte dat)
{
//	if ((!(g_regs.IE&INT_SERIAL))||(g_regs.IF&INT_SERIAL))
//		return 0xFF;

	if ((g_regs.SC&0x81)==0x80){
//		fprintf(file,"seri_recv my:%02X tar:%02X state SC:%02X\n",g_regs.SB,dat,g_regs.SC);
		byte ret=g_regs.SB;
		g_regs.SB=dat;
		g_regs.SC&=1;
		cpu_irq(INT_SERIAL);
		return ret;
	}
	else
		return 0xFF;
}

void cpu_irq(int irq_type)
{
//	fprintf(file,"irq %02X LCDC %02X\n",irq_type,g_regs.LCDC);
	if (!((irq_type==INT_VBLANK||irq_type==INT_LCDC)&&(!(g_regs.LCDC&0x80))))
//		if (last_int!=irq_type)
			g_regs.IF|=(irq_type/*&g_regs.IE*/);
//	g_regs.IF|=irq_type;
	cpu_irq_check();
}

//g_regs.IF, g_regs.IE, c_regs_I, halt, int_disable_next ύXꂽɌĂԂƁBŃtOāAC[vŌ - LCK
void cpu_irq_check()
{
	int_invoke_next=((g_regs.IF&g_regs.IE)&&(c_regs_I||halt) || int_disable_next);	//肱݂鎞+disable_next
}

//荞݂ώZclock悤ɕύX
void cpu_irq_process()
{
	if (int_disable_next){
		int_disable_next=false;
		cpu_irq_check();
		return;
	}

	if (!((g_regs.IF&g_regs.IE)&&(c_regs_I||halt))) {
		cpu_irq_check();
		return;
	}

	if (halt)
		c_regs_PC++;

	write(c_regs_SP-2,c_regs_PC&0xFF);
	write(c_regs_SP-1,(c_regs_PC>>8));
	c_regs_SP-=2;
	
	if (g_regs.IF&g_regs.IE&INT_VBLANK){//VBlank
		c_regs_PC=0x40;
		g_regs.IF&=0xFE;
		last_int=INT_VBLANK;
	}
	else if (g_regs.IF&g_regs.IE&INT_LCDC){//LCDC
		c_regs_PC=0x48;
		g_regs.IF&=0xFD;
		last_int=INT_LCDC;
	}
	else if (g_regs.IF&g_regs.IE&INT_TIMER){//Timer
		c_regs_PC=0x50;
		g_regs.IF&=0xFB;
		last_int=INT_TIMER;
	}
	else if (g_regs.IF&g_regs.IE&INT_SERIAL){//Serial
		c_regs_PC=0x58;
		g_regs.IF&=0xF7;
		last_int=INT_SERIAL;
	}
	else if (g_regs.IF&g_regs.IE&INT_PAD){//Pad
		c_regs_PC=0x60;
		g_regs.IF&=0xEF;
		last_int=INT_PAD;
	}
	else {}

	halt=false;
	c_regs_I=0;
//		g_regs.IF=0;

	cpu_irq_check();
	
	//荞݂̂ŃXe[g𑫂
	//13state͎blBGBłǂȂĂ邩͒mȂB
	//悭킩̂sys_clock͑ĂȂĂ - LCK
//	rest_clock-=13;
//	div_clock+=13;
//	total_clock+=13;

	return;
}

int tmp_clocks;
#include "op_cb.h"
#include "op_normal.h"

void cpu_exec(int clocks)
{
	int op_code;
	static const int timer_clocks[]={1024,16,64,256};

	if (speed)
		clocks*=2;

	rp_que[0]=clocks+8;
	rp_que[1]=0x00000000;
	que_cur=1;

	rest_clock+=clocks;

	if (gdma_rest){
		if (rest_clock<=gdma_rest){
			gdma_rest-=rest_clock;
			sys_clock+=rest_clock;
			div_clock+=rest_clock;
			total_clock+=rest_clock;
			rest_clock=0;
		}
		else{
			rest_clock-=gdma_rest;
			sys_clock+=gdma_rest;
			div_clock+=gdma_rest;
			total_clock+=gdma_rest;
			gdma_rest=0;
		}
	}

	while(rest_clock>0){
		if (int_invoke_next) cpu_irq_process();

//		printf("PC=%d\n", c_regs.PC);

		op_code=op_read();
		tmp_clocks=cycles[op_code];

//		if (b_trace)
//			log();

		op_code_normal_JmpTbl[op_code]();

		rest_clock-=tmp_clocks;
		div_clock+=tmp_clocks;
		total_clock+=tmp_clocks;

		if (g_regs.TAC&0x04){//^C}肱
			sys_clock+=tmp_clocks;
			if (sys_clock>timer_clocks[g_regs.TAC&0x03]){
				sys_clock&=timer_clocks[g_regs.TAC&0x03]-1;
				g_regs.TIMA++;
				if (!g_regs.TIMA){
					cpu_irq(INT_TIMER);
					g_regs.TIMA=g_regs.TMA;
				}
			}
		}

		if (div_clock&0x100){
			g_regs.DIV-=div_clock>>8;
			div_clock&=0xff;
		}

		if (total_clock>seri_occer){
			seri_occer=0x7fffffff;
/*			if (ref_gb->get_target()){
				byte ret=ref_gb->get_target()->get_cpu()->seri_send(g_regs.SB);
				g_regs.SB=ret;
				g_regs.SC&=3;
			}
			else*/{
				if (hook_ext){ // tbN܂
					byte ret=hook_proc.send(g_regs.SB);
					g_regs.SB=ret;
					g_regs.SC&=3;
				}
				else{
					g_regs.SB=0xff;
					g_regs.SC&=3;
				}
			}
			cpu_irq(INT_SERIAL);
		}
	}
}

void cpu_set_trace(bool trace) { b_trace=trace; }
byte *cpu_get_vram() { return vram; }
byte *cpu_get_ram() { return ram; }
byte *cpu_get_oam() { return oam; }
byte *cpu_get_stack() { return stack; }
byte *cpu_get_ram_bank() { return ram_bank; }
void cpu_set_ram_bank(int bank) { ram_bank=ram+bank*0x1000; }
//struct cpu_regs *cpu_get_regs() { return &c_regs; }
int cpu_get_clock() { return total_clock; }
bool cpu_get_speed() { return speed; }
bool *cpu_get_halt() { return &halt; }

struct cpu_regs *cpu_get_c_regs() 
{ 
	_c_regs.AF.w=c_regs_AF;
	_c_regs.BC.w=c_regs_BC;
	_c_regs.DE.w=c_regs_DE;
	_c_regs.HL.w=c_regs_HL;
	_c_regs.SP  =c_regs_SP;
	_c_regs.PC  =c_regs_PC;
	_c_regs.I   =c_regs_I ;
	return &_c_regs; 
}

struct cpu_regs *cpu_set_c_regs() 
{ 
	c_regs_AF=_c_regs.AF.w;
	c_regs_BC=_c_regs.BC.w;
	c_regs_DE=_c_regs.DE.w;
	c_regs_HL=_c_regs.HL.w;
	c_regs_SP=_c_regs.SP  ;
	c_regs_PC=_c_regs.PC  ;
	c_regs_I =_c_regs.I   ;
	cpu_irq_check(); 
	return &_c_regs; 
}

