#include "main.h"

int save_sram(byte *buf,int size)
{
	int fd, tmp;
	int sram_tbl[]={1,1,1,4,16,8};

	fd = sceIoOpen(SavePath,O_CREAT|O_WRONLY|O_TRUNC, 0777);
	if(fd>=0){
		sceIoWrite(fd, buf, 0x2000*sram_tbl[size]);
		if ((rom_get_info()->cart_type>=0x0f)&&(rom_get_info()->cart_type<=0x13)){
			tmp=renderer_get_timer_state();
			sceIoWrite(fd, &tmp, 4);
		}
		sceIoClose(fd);
		return 1;
	}else
		return 0;
}

int load_sram(void)
{
	int fd;
	int ramsize;
	char *p;
	
	_strcpy(SavePath, RomPath);
	p = _strrchr(SavePath, '.');
	_strcpy(p, ".sav");
	fd=sceIoOpen(SavePath, O_RDONLY, 0777);
	
	if(fd<0){
		_strcpy(SavePath,RinPath);
		_strcat(SavePath,"SAVE/");
		_strcat(SavePath,RomName);
		_strcat(SavePath, ".sav");
		fd=sceIoOpen(SavePath, O_RDONLY, 0777);
		if(fd<0) return 0;
	}
	
	ramsize = sceIoRead(fd, sram_space, 16*0x2000+4);
	if(ramsize & 4)
		renderer_set_timer_state(*(int*)(sram_space+ramsize-4));
	sceIoClose(fd);
	return ramsize;
}

// PackBitsk
int PackBits_encode( byte *in, byte *out, int size )
{
	byte ltr[128], bak, cur, *wp=out;
	int n=1, m=0, i=0, j;
	
	bak = in[i++];
	while(i<size){
		cur = in[i++];
		if(cur != bak){
			if(n>=2){
				*wp++ = -(n-1);
				*wp++ = bak;
			}else{
				ltr[m++] = bak;
				if(m >= 128){
					*wp++ = m-1;
					for(j=0; j<m; j++) *wp++ = ltr[j];
					m=0;
				}
			}
			bak = cur;
			n = 1;
		}else{
			if(m >= 1){
				*wp++ = m-1;
				for(j=0; j<m; j++) *wp++ = ltr[j];
				m = 0;
			}
			if(n < 128){
				n++;
			}else{
				*wp++ = -(n-1);
				*wp++ = bak;
				n = 1;
			}
		}
	}
	if(n>=2){
		if(m >= 1){
			*wp++ = m-1;
			for(j=0; j<m; j++) *wp++ = ltr[j];
		}
		*wp++ = -(n-1);
		*wp++ = bak;
	}else{
		ltr[m++] = bak;
		*wp++ = m-1;
		for(j=0; j<m; j++) *wp++ = ltr[j];
	}
	
	return wp - out;
}

// PackBitsk
int PackBits_decode( byte *in, byte *out, int size )
{
	byte *limit=in+size, *wp=out;
	int i, c;
	
	while( in < limit ){
		c = (char)(*in++);
		if(c < 0){
			for(i=1-c; i>0; i--) *wp++ = *in;
			in++;
		}else{
			for(i=1+c; i>0; i--) *wp++ = *in++;
		}
	}
	
	return wp - out;
}

void get_state_path(int slot, char *out)
{
	char *p, ext[5];
	
	_strcpy(ext, ".sv0");
	ext[3]=slot+'0';
	_strcpy(out, SavePath);
	p = _strrchr(out, '.');
	_strcpy(p, ext);
}

void get_thumb_path(int slot, char *out)
{
	char *p, ext[5];
	
	_strcpy(ext, ".tn0");
	ext[3]=slot+'0';
	_strcpy(out, SavePath);
	p = _strrchr(out, '.');
	_strcpy(p, ext);
}

int delete_thumb(int slot)
{
	int ret;
	char path[MAXPATH];
	
	get_thumb_path(slot,path);
	ret = sceIoRemove(path);
	
	return ret;
}

int delete_state(int slot)
{
	int ret;
	char path[MAXPATH];
	
	get_state_path(slot,path);
	ret = sceIoRemove(path);
	if(ret<0) return ret;
	
	delete_thumb(slot);
	
	return ret;
}

int save_thumb(int slot)
{
	int fd;
	char path[MAXPATH];
	
	get_thumb_path(slot,path);
	fd = sceIoOpen(path,O_CREAT|O_WRONLY|O_TRUNC, 0777);
	if(fd<0) return 0;
	
	word bgr, r, g, b;
	int i, j;
	
	// F
	for(i=0; i<144; i++){
		for(j=0; j<160; j++){
			bgr = vframe[i*176+8+j];
			r = (bgr    )&0x18;
			g = (bgr>>5 )&0x18;
			b = (bgr>>10)&0x18;
			thumb_b[i*160+j] = (b<<1) | (g>>1) | (r>>3);
		}
	}
	
	// PackBitsk
	int size = PackBits_encode( thumb_b, (byte*)thumb_w, 160*144 );
	
	int ret = sceIoWrite(fd, thumb_w, size);
	sceIoClose(fd);
	
	if(ret!=size){
		delete_thumb(slot);
		return 0;
	}else
		return 1;
}

int load_thumb(int slot)
{
	int fd, size;
	char path[MAXPATH];
	
	get_thumb_path(slot,path);
	fd = sceIoOpen(path,O_RDONLY, 0777);
	if(fd<0) return 0;
	size = sceIoRead(fd, thumb_w, sizeof(thumb_w));
	sceIoClose(fd);
	
	PackBits_decode((byte*)thumb_w, thumb_b, size);
	
	byte bgr;
	word r, g, b;
	int i;
	
	for(i=0; i<160*144; i++){
		bgr = thumb_b[i];
		r = (bgr   )&0x3;
		g = (bgr>>2)&0x3;
		b = (bgr>>4)&0x3;
		thumb_w[i] = (b<<13) | (g<<8) | (r<<3);
	}
	
	return 1;
}

int save_state(int slot)
{
	int fd, ret;
	char path[MAXPATH];
	
	get_state_path(slot,path);
	fd = sceIoOpen(path,O_CREAT|O_WRONLY|O_TRUNC, 0777);
	if(fd<0) return 0;
	gb_save_state(fd);
	sceIoClose(fd);
	
	if(setting.thumb)
		ret = save_thumb(slot);
	else{
		ret = 1;
		delete_thumb(slot);
	}
	
	return ret;
}

int load_state(int slot){
	int fd;
	char path[MAXPATH];
	
	get_state_path(slot,path);
	fd = sceIoOpen(path,O_RDONLY, 0777);
	if(fd<0) return 0;
	gb_restore_state(fd);
	sceIoClose(fd);
	
	if(rom_get_info()->gb_type == 1)
		lcd_set_mpal(0);
	else if(rom_get_info()->gb_type == 2 && sgb_mode)
		lcd_set_mpal(PAL_SGB);
	else
		lcd_set_mpal(setting.gb_palette);
		
	if(rom_get_info()->gb_type>=3 && org_gbtype==3)
		now_gb_mode = 3;
	else if(rom_get_info()->gb_type==2 && sgb_mode)
		now_gb_mode = 2;
	else
		now_gb_mode = 1;
	
	border_uploaded = 2;
	
	return 1;
}

